(*************************************************************
   2003 yeahware software development
  http://www.yeahware.com

  If this code works, it was written by Markus Fuchs.
  If not, I don't know who wrote it.  (Paul DiLascia)

  Module           :  enable
  Source File Name :  uenable.pas
  Description      :  Delphi translation of NTDDK example
  Compiler         :  Delphi 5
  Platforms        :  Windows 9x and NT platform
  Created          :  24.09.2003
  Revisions        :  21.10.2003 - Implemented Win9x support
                      11.11.2003 - Improved TreeView handling
*************************************************************)

unit uenable;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ImgList, ComCtrls, StdCtrls, SetupAPI, ExtCtrls, Menus;


const
  SetupApiModuleName        = 'cfgmgr32.dll';
  REGSTR_VAL_NODISPLAYCLASS = 'NoDisplayClass';
  CR_SUCCESS                = $00000000;
  DN_HAS_PROBLEM            = $00000400;
  DN_DISABLEABLE            = $00002000;
  DN_NO_SHOW_IN_DM          = $40000000;
  CM_PROB_DISABLED          = $00000016;
  CM_PROB_HARDWARE_DISABLED = $0000001D;

  function CM_Get_DevNode_Status(pulStatus: PULong; pulProblemNumber: PULong;
  dnDevInst: DWord; ulFlags: ULong): DWord; stdcall;
  external SetupApiModuleName name 'CM_Get_DevNode_Status';

type
  TForm1 = class(TForm)
    TreeView: TTreeView;
    ImageList: TImageList;
    MainMenu: TMainMenu;
    mFile: TMenuItem;
    mExit: TMenuItem;
    mChange: TMenuItem;
    mEnableDevice: TMenuItem;
    mDisableDevice: TMenuItem;
    mOptions: TMenuItem;
    mRefreshDisplay: TMenuItem;
    mShowHiddenDevices: TMenuItem;
    mN: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure mExitClick(Sender: TObject);
    procedure mRefreshDisplayClick(Sender: TObject);
    procedure mShowHiddenDevicesClick(Sender: TObject);
    procedure TreeViewChange(Sender: TObject; Node: TTreeNode);
    procedure TreeViewDblClick(Sender: TObject);
    procedure ChangeEnableDevice(Sender: TObject);
    procedure ChangeDisableDevice(Sender: TObject);
  private
    { Private-Deklarationen }
    DevInfo: hDevInfo;
    ClassImageListData: TSPClassImageListData;
    ShowHidden: Boolean;
    function IsDisableable(SelectedItem: DWord; hDevInfo: hDevInfo): Boolean;
    function IsDisabled(SelectedItem: DWord; hDevInfo: hDevInfo): Boolean;
    function StateChange(NewState: DWord; SelectedItem: DWord; hDevInfo: hDevInfo): Boolean;
    function GetRegistryProperty(PnPHandle: HDEVINFO;DevData: TSPDevInfoData;
             Prop: DWORD; Buffer: PChar; dwLength: DWord): Boolean;
    function ConstructDeviceName(DeviceInfoSet: hDevInfo;
             DeviceInfoData: TSPDevInfoData; Buffer: PChar; dwLength: DWord): Boolean;
    function IsClassHidden(ClassGuid: TGuid): Boolean;
    function EnumAddDevices(ShowHidden: Boolean; hwndTree: TTreeView; DevInfo: hDevInfo): Boolean;
    function GetClassImageIndex(ClassGuid: TGuid; Index: PInt): Boolean;
  public
    { Public-Deklarationen }
  end;

var
  Form1: TForm1;


implementation

{$R *.DFM}

function TForm1.IsDisableable(SelectedItem: DWord; hDevInfo: hDevInfo): Boolean;
var
  DeviceInfoData: TSPDevInfoData;
  Status, Problem: DWord;
begin
  DeviceInfoData.cbSize := SizeOf(TSPDevInfoData);
  // Get a handle to the Selected Item.
  if (not SetupDiEnumDeviceInfo(hDevInfo, SelectedItem, DeviceInfoData)) then
  begin
    Result := false;
    exit;
  end;
  if (CM_Get_DevNode_Status(@Status, @Problem, DeviceInfoData.DevInst, 0) <> CR_SUCCESS) then
  begin
    Result := false;
    exit;
  end;
  Result := ((Status and DN_DISABLEABLE = DN_DISABLEABLE) and not (CM_PROB_HARDWARE_DISABLED = Problem));
end;

function TForm1.IsDisabled(SelectedItem: DWord; hDevInfo: hDevInfo): Boolean;
var
  DeviceInfoData: TSPDevInfoData;
  Status, Problem: DWord;
begin
  DeviceInfoData.cbSize := SizeOf(TSPDevInfoData);
  // Get a handle to the Selected Item.
  if (not SetupDiEnumDeviceInfo(hDevInfo, SelectedItem, DeviceInfoData)) then
  begin
    Result := false;
    exit;
  end;
  if (CM_Get_DevNode_Status(@Status, @Problem, DeviceInfoData.DevInst, 0) <> CR_SUCCESS) then
  begin
    Result := false;
    exit;
  end;
  Result := ((Status and DN_HAS_PROBLEM = DN_HAS_PROBLEM) and (CM_PROB_DISABLED = Problem));
end;

function TForm1.StateChange(NewState: DWord; SelectedItem: DWord; hDevInfo: hDevInfo): Boolean;
var
  PropChangeParams: TSPPropChangeParams;
  DeviceInfoData: TSPDevInfoData;
begin
  PropChangeParams.ClassInstallHeader.cbSize := SizeOf(TSPClassInstallHeader);
  DeviceInfoData.cbSize := SizeOf(TSPDevInfoData);
  // Get a handle to the Selected Item.
  if (not SetupDiEnumDeviceInfo(hDevInfo, SelectedItem, DeviceInfoData)) then
  begin
    Result := false;
    ShowMessage('EnumDeviceInfo');
    exit;
  end;
  // Set the PropChangeParams structure.
  PropChangeParams.ClassInstallHeader.InstallFunction := DIF_PROPERTYCHANGE;
  PropChangeParams.Scope := DICS_FLAG_GLOBAL;
  PropChangeParams.StateChange := NewState;
  if (not SetupDiSetClassInstallParams(hDevInfo, @DeviceInfoData,
     PSPClassInstallHeader(@PropChangeParams), SizeOf(PropChangeParams))) then
  begin
    Result := false;
    ShowMessage('SetClassInstallParams');
    exit;
  end;
  // Call the ClassInstaller and perform the change.
  if (not SetupDiCallClassInstaller(DIF_PROPERTYCHANGE, hDevInfo, @DeviceInfoData)) then
  begin
    Result := false;
    ShowMessage('SetClassInstallParams');
    exit;
  end;
  Result := true;
end;

function TForm1.GetClassImageIndex(ClassGuid: TGuid; Index: PInt): Boolean;
begin
  Result := SetupDiGetClassImageIndex(ClassImageListData, ClassGuid, Index^);
end;

function TForm1.GetRegistryProperty(PnPHandle: hDevInfo; DevData: TSPDevInfoData; Prop: DWord; Buffer: PChar; dwLength: DWord): Boolean;
var
  aBuffer: array[0..256] of Char;
begin
  dwLength := 0;
  aBuffer[0] := #0;
  SetupDiGetDeviceRegistryProperty(PnPHandle, DevData, Prop, Prop, PBYTE(@aBuffer[0]), SizeOf(aBuffer), dwLength);
  StrCopy(Buffer, aBuffer);
  Result := Buffer^ <> #0;
end;

function TForm1.ConstructDeviceName(DeviceInfoSet: hDevInfo;
         DeviceInfoData: TSPDevInfoData; Buffer: PChar; dwLength: DWord): Boolean;
const
  UnknownDevice = '<Unknown Device>';
begin
  if (not GetRegistryProperty(DeviceInfoSet, DeviceInfoData, SPDRP_FRIENDLYNAME, Buffer, dwLength)) then
  begin
    if (not GetRegistryProperty(DeviceInfoSet, DeviceInfoData, SPDRP_DEVICEDESC, Buffer, dwLength)) then
    begin
      if (not GetRegistryProperty(DeviceInfoSet, DeviceInfoData, SPDRP_CLASS, Buffer, dwLength)) then
      begin
        if (not GetRegistryProperty(DeviceInfoSet, DeviceInfoData, SPDRP_CLASSGUID, Buffer, dwLength)) then
        begin
          dwLength := DWord(SizeOf(UnknownDevice));
          Buffer := Pointer(LocalAlloc(LPTR, Cardinal(dwLength)));
          StrCopy(Buffer, UnknownDevice);
        end;
      end;
    end;
  end;
  Result := true;
end;

function TForm1.IsClassHidden(ClassGuid: TGuid): Boolean;
var
  bHidden: Boolean;
  hKeyClass: HKey;
begin
  bHidden := false;
  hKeyClass := SetupDiOpenClassRegKey(@ClassGuid, KEY_READ);
  if (hKeyClass <> 0) then
  begin
    bHidden := (RegQueryValueEx(hKeyClass, REGSTR_VAL_NODISPLAYCLASS, nil, nil, nil, nil) = ERROR_SUCCESS);
    RegCloseKey(hKeyClass);
  end;
  Result := bHidden;
end;

function TForm1.EnumAddDevices(ShowHidden: Boolean; hwndTree: TTreeView; DevInfo: hDevInfo): Boolean;
var
  i, Status, Problem: DWord;
  pszText: PChar;
  DeviceInfoData: TSPDevInfoData;
  iImage: Integer;
begin
  TTreeView(hWndTree).Items.BeginUpdate;
  DeviceInfoData.cbSize := SizeOf(TSPDevInfoData);
  // Clean off all the items in a TreeView.
  TTreeView(hWndTree).Items.Clear;
  i := 0;
  // Enumerate though all the devices.
  while SetupDiEnumDeviceInfo(DevInfo, i, DeviceInfoData) do
  begin
    inc(i);
    // Should we display this device, or move onto the next one.
    if (CM_Get_DevNode_Status(@Status, @Problem, DeviceInfoData.DevInst, 0) <> CR_SUCCESS) then
    begin
      ShowMessage('CM_Get_DevNode_Status :: uenable.pas');
      break;
    end;
    if (not (ShowHidden or not(Boolean(Status and DN_NO_SHOW_IN_DM) or IsClassHidden(DeviceInfoData.ClassGuid)))) then
    begin
      break;
    end;
    GetMem(pszText, 256);
    try
      // Get a friendly name for the device.
      ConstructDeviceName(DevInfo, DeviceInfoData, pszText, DWord(nil));
      // Try to get an icon index for this device.
      if (GetClassImageIndex(DeviceInfoData.ClassGuid, @iImage)) then
      begin
        with TTreeView(hWndTree).Items.AddObject(nil, pszText, nil) do
        begin
          TTreeView(hWndTree).Items[i-1].ImageIndex := iImage;
          TTreeView(hWndTree).Items[i-1].SelectedIndex := iImage;
        end;
        if (Problem = CM_PROB_DISABLED) then // red (X)
        begin
            TTreeView(hWndTree).Items[i-1].OverlayIndex := IDI_DISABLED_OVL - IDI_CLASSICON_OVERLAYFIRST;
        end else
        begin
          if (Boolean(Problem)) then // yellow (!)
          begin
              TTreeView(hWndTree).Items[i-1].OverlayIndex := IDI_PROBLEM_OVL - IDI_CLASSICON_OVERLAYFIRST;
          end;
        end;
        if (Status and DN_NO_SHOW_IN_DM = DN_NO_SHOW_IN_DM) then // Greyed out
        begin
          TTreeView(hWndTree).Items[i-1].Cut := true;
        end;
      end;
    finally
      FreeMem(pszText);
    end;
  end;
  TTreeView(hWndTree).Items.EndUpdate;
  Result := true;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  if (not LoadSetupAPI) then
  begin
    ShowMessage('Could not load SetupAPI.dll');
    exit;
  end;
  DevInfo := nil;
  ShowHidden := false;
  // Get a handle to all devices in all classes present on system
  DevInfo := SetupDiGetClassDevs(nil, nil, 0, DIGCF_PRESENT or DIGCF_ALLCLASSES);
  if (DevInfo = Pointer(INVALID_HANDLE_VALUE)) then
  begin
    ShowMessage('GetClassDevs');
    exit;
  end;
  // Get the Images for all classes, and bind to the TreeView
  ClassImageListData.cbSize := SizeOf(TSPClassImageListData);
  if (not SetupDiGetClassImageList(ClassImageListData)) then
  begin
    ShowMessage('GetClassImageList');
    exit;
  end;
  ImageList.Handle := ClassImageListData.ImageList;
  TreeView.Images := ImageList;
  // Add the devices to the TreeView window.
  EnumAddDevices(ShowHidden, TreeView, DevInfo);
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  SetupDiDestroyDeviceInfoList(DevInfo);
  SetupDiDestroyClassImageList(ClassImageListData);
  UnloadSetupApi;
end;

procedure TForm1.mExitClick(Sender: TObject);
begin
  Close;
end;

procedure TForm1.mRefreshDisplayClick(Sender: TObject);
begin
  EnumAddDevices(ShowHidden, TreeView, DevInfo);
end;

procedure TForm1.mShowHiddenDevicesClick(Sender: TObject);
begin
  ShowHidden := not ShowHidden;
  MainMenu.Items[2][2].Checked := ShowHidden;
  EnumAddDevices(ShowHidden, TreeView, DevInfo);
end;

procedure TForm1.TreeViewChange(Sender: TObject; Node: TTreeNode);
begin
  with MainMenu, TreeView.Selected do
  begin
    // Clear the Menu Items.
    Items[1][0].enabled := false;
    Items[1][1].enabled := false;
    if (IsDisabled(Index, DevInfo)) then
    begin
      Items[1][0].enabled := true
    end else
    begin
      if (IsDisableable(Index, DevInfo)) then Items[1][1].enabled := true
    end;
  end;
end;

procedure TForm1.TreeViewDblClick(Sender: TObject);
begin
  with TreeView.Selected do
  begin
    if (IsDisabled(Index, DevInfo)) then
    begin
      ChangeEnableDevice(Self);;
    end else
    begin
      if (IsDisableable(Index, DevInfo)) then ChangeDisableDevice(Self);
    end;
  end;
end;

procedure TForm1.ChangeEnableDevice(Sender: TObject);
begin
  if (MessageBox(Handle, 'Enable this Device?', 'Change Device Status', MB_OKCANCEL) = IDOK) then
  begin
    if (StateChange(DICS_ENABLE, TreeView.Selected.Index, DevInfo)) then EnumAddDevices(ShowHidden, TreeView, DevInfo);
  end;
end;

procedure TForm1.ChangeDisableDevice(Sender: TObject);
begin
  if (MessageBox(Handle, 'Disable this Device?', 'Change Device Status', MB_OKCANCEL) = IDOK) then
  begin
    if (StateChange(DICS_DISABLE, TreeView.Selected.Index, DevInfo)) then EnumAddDevices(ShowHidden, TreeView, DevInfo);
  end;
end;

end.
